# StrategicGames.jl
A set of functions in pure Julia for analysing strategic games using the concepts and tools of the Game Theory

In particular, `nash_lcp(payoff_tensor;init,verbosity)` implements the LCP algorithm to find a Nash Equilibrium for n-players simultaneous games when mixed strategies are allowed.

While written in Julia, the library is easily accessible in R or Python with [`JuliaCall`](https://github.com/Non-Contradiction/JuliaCall) and [`PyJulia`](https://github.com/JuliaPy/pyjulia) packages respectively.


[![](https://img.shields.io/badge/docs-stable-blue.svg)](https://sylvaticus.github.io/StrategicGames.jl/stable)
[![](https://img.shields.io/badge/docs-dev-blue.svg)](https://sylvaticus.github.io/StrategicGames.jl/dev)
[![Build status (Github Actions)](https://github.com/sylvaticus/StrategicGames.jl/workflows/CI/badge.svg)](https://github.com/sylvaticus/StrategicGames.jl/actions)
[![codecov.io](http://codecov.io/github/sylvaticus/StrategicGames.jl/coverage.svg?branch=main)](http://codecov.io/github/sylvaticus/StrategicGames.jl?branch=main)


## Examples

```julia
julia> using Pkg; Pkg.add("StrategicGames") # only once to install the library
julia> using StrategicGames
```

### 3-players, 2 options per player with payoff depending on number of players choosing the same option

The 3 players in the game can choose between `X` or `Y` with payoffs respectively `2nₓ - 2nₓ²+3` and `4-nᵧ` where `nₓ` and `nᵧ` represent the number of players that choose `X` or `Y`.
(This example is taken from https://www.youtube.com/watch?v=bKrwQKUT0v8 where it is analytically solved)
```julia
julia> U = [(0,0,0) ; (3,3,3) ;; (3,3,3)   ; (2,2,4) ;;;
            (3,3,3)  ; (2,4,2)  ;; (4,2,2) ; (1,1,1)  ;;;]
2×2×2 Array{Tuple{Int64, Int64, Int64}, 3}:
[:, :, 1] =
 (0, 0, 0)  (3, 3, 3)
 (3, 3, 3)  (2, 2, 4)

[:, :, 2] =
 (3, 3, 3)  (4, 2, 2)
 (2, 4, 2)  (1, 1, 1)

julia> payoff_array = expand_dimensions(U) # from n-dimensional array of tuples to n+1 arrays of scalars
2×2×2×3 Array{Int64, 4}:
[:, :, 1, 1] =
 0  3
 3  2

[:, :, 2, 1] =
 3  4
 2  1

[:, :, 1, 2] =
 0  3
 3  2

[:, :, 2, 2] =
 3  2
 4  1

[:, :, 1, 3] =
 0  3
 3  4

[:, :, 2, 3] =
 3  2
 2  1

julia> eq  = nash_lcp(payoff_array)
(status = MathOptInterface.LOCALLY_SOLVED, equilibrium_strategies = [[0.5811388300841898, 0.4188611699158103], [0.5811388300841898, 0.4188611699158103], [0.5811388300841898, 0.41886116991581035]], expected_payoffs = [2.16227766016838, 2.16227766016838, 2.16227766016838])

julia> eq_strategies = eq.equilibrium_strategies
3-element Vector{Vector{Float64}}:
 [0.5811388300841898, 0.4188611699158103]
 [0.5811388300841898, 0.4188611699158103]
 [0.5811388300841898, 0.41886116991581035]

julia> p = -1 + sqrt(10)/2
0.5811388300841898

julia> @test eq_strategies ≈ [[p,1-p],[p,1-p],[p,1-p]]
Test Passed

julia> expected_payoff(payoff_array,eq_strategies)
3-element Vector{Float64}:
 2.1622776601683795
 2.1622776601683795
 2.1622776601683795
 ```

### Prisoner's dilemma
 ```julia
julia> payoff = [(-1,-1) (-3,0); (0, -3) (-2, -2)]
2×2 Matrix{Tuple{Int64, Int64}}:
 (-1, -1)  (-3, 0)
 (0, -3)   (-2, -2)
julia> eq     = nash_lcp(expand_dimensions(payoff));
julia> eq_strategies = eq.equilibrium_strategies
2-element Vector{Vector{Float64}}:
 [-4.049752569180346e-11, 1.0000000000404976]
 [-4.0497525691839856e-11, 1.0000000000404976]
```



 ## Other libraries

 - [Nash.jl](https://github.com/KrainskiL/Nash.jl)
 - [GameTheory.jl](https://github.com/QuantEcon/GameTheory.jl)
 - [Nashpy](https://github.com/drvinceknight/Nashpy): two players only
 - [Gambit](http://www.gambit-project.org/)
